/*
 *  Copyright 2012-2025 Carnegie Mellon University
 *  See license information in LICENSE.txt.
 */
/*
 *  mediator_dedup.c
 *
 *  General deduplication code.
 *
 *  ------------------------------------------------------------------------
 *  Authors: Emily Sarneso
 *  ------------------------------------------------------------------------
 *  @DISTRIBUTION_STATEMENT_BEGIN@
 *  super_mediator-1.11
 *
 *  Copyright 2024 Carnegie Mellon University.
 *
 *  NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
 *  INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
 *  UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED,
 *  AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR
 *  PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS OBTAINED FROM USE OF
 *  THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF
 *  ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT
 *  INFRINGEMENT.
 *
 *  Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
 *  contact permission@sei.cmu.edu for full terms.
 *
 *  [DISTRIBUTION STATEMENT A] This material has been approved for public
 *  release and unlimited distribution.  Please see Copyright notice for
 *  non-US Government use and distribution.
 *
 *  This Software includes and/or makes use of Third-Party Software each
 *  subject to its own license.
 *
 *  DM24-1038
 *  @DISTRIBUTION_STATEMENT_END@
 *  ------------------------------------------------------------------------
 */

#include "mediator_dedup.h"
#include "mediator_core.h"
#include "mediator_inf.h"
#include "mediator_print.h"

/* Default hit count */
#define MD_DEDUP_DEFAULT_HIT    5000

/* Default flush timeout in seconds */
#define MD_DEDUP_DEFAULT_FLUSH  300

/* fixbuf 2.x uses char* as the type of the name of info elements in
 * fbInfoElementSpec_t; wrap this around string literals to quiet compiler
 * warnings */
#define C(String) (char *)String

#define DEDUP_DEBUG 0

/* note that unlike others this is by positional index */
#define FBBLNEXT(a, b) fbBasicListGetIndexedDataPtr((a), (b))

#define SSL_SERIAL_IE 244
#define SSL_COMMON_NAME 3
#define SSL_ORG_UNIT 11

#define MD_APPEND_CHAR(_buf_, _ch_) \
    *(_buf_->cp) = _ch_;            \
    ++(_buf_->cp);



typedef struct md_dedup_str_node_st md_dedup_str_node_t;
struct md_dedup_str_node_st {
    md_dedup_str_node_t  *next;
    md_dedup_str_node_t  *prev;
    yfTime_t              ftime;
    yfTime_t              ltime;
    uint64_t              hitcount;
    yfTime_t              stime;
    uint32_t              hash;
    uint16_t              ie;
    size_t                caplen;
    uint8_t              *data;
};

typedef struct md_dedup_stats_st {
    uint64_t   recvd;
    uint64_t   flushed;
} md_dedup_stats_t;

typedef struct md_dedup_node_st md_dedup_node_t;
struct md_dedup_node_st {
    md_dedup_node_t      *next;
    md_dedup_node_t      *prev;
    md_dedup_str_node_t  *strnode;
    md_dedup_ie_t        *ietab;
    md_dedup_t            exnode;
};

typedef struct md_dedup_cqueue_st {
    md_dedup_node_t  *head;
    md_dedup_node_t  *tail;
} md_dedup_cqueue_t;

typedef struct md_dedup_ssl_node_st {
    uint8_t   *serial;
    size_t     serial_len;
    uint8_t   *issuer;
    size_t     issuer_len;
    uint64_t   count;
} md_dedup_ssl_node_t;

typedef struct mdMapKey4_st {
    uint32_t   ip;
    uint32_t   map;
} mdMapKey4_t;

typedef struct mdMapKey6_st {
    uint8_t    ip[16];
    uint32_t   map;
} mdMapKey6_t;

typedef struct md_dedup_tmpl_st md_dedup_tmpl_t;

struct md_dedup_tmpl_st {
    fbTemplate_t  *tmpl;
    uint16_t       intid;
    uint16_t       extid;
    uint16_t       ie;
};

struct md_dedup_state_st {
    GHashTable         *ie_table;
    GHashTable         *cert_table;
    md_dedup_cqueue_t  *cq;
    md_dedup_ie_t      *head;
    md_dedup_ie_t      *tail;
    md_dedup_stats_t    stats;
    yfDiffTime_t        flush_timeout;
    uint64_t            max_hit_count;
    gboolean            merge;
    gboolean            add_export;
};

typedef struct md_dedup_ip_node_st md_dedup_ip_node_t;
struct md_dedup_ip_node_st {
    md_dedup_ip_node_t   *next;
    md_dedup_ip_node_t   *prev;
    md_dedup_str_node_t  *head;
    md_dedup_str_node_t  *tail;
    smFieldMap_t         *map;
    /*smVarHashKey_t      *sip6_key;*/
    mdMapKey6_t          *sip6_key;
    mdMapKey4_t          *sip_key;
};

typedef struct md_dedup_ssl_str_node_st md_dedup_ssl_str_node_t;
struct md_dedup_ssl_str_node_st {
    md_dedup_ssl_str_node_t  *next;
    md_dedup_ssl_str_node_t  *prev;
    yfTime_t                  ftime;
    yfTime_t                  ltime;
    uint64_t                  hitcount;
    yfTime_t                  stime;
    uint32_t                  hash;
    md_dedup_ssl_node_t      *cert1;
    md_dedup_ssl_node_t      *cert2;
};

typedef struct md_dedup_ssl_ip_node_st md_dedup_ssl_ip_node_t;
struct md_dedup_ssl_ip_node_st {
    md_dedup_ssl_ip_node_t   *next;
    md_dedup_ssl_ip_node_t   *prev;
    md_dedup_ssl_str_node_t  *head;
    md_dedup_ssl_str_node_t  *tail;
    smFieldMap_t             *map;
    /*    smVarHashKey_t         *sip6_key;*/
    mdMapKey6_t              *sip6_key;
    mdMapKey4_t              *sip_key;
};

/* used in parser */
struct md_dedup_ie_st {
    md_dedup_ie_t       *next;
    md_dedup_ie_t       *prev;
    md_dedup_ip_node_t  *head;
    md_dedup_ip_node_t  *tail;
    smHashTable_t       *ip_table;
    smHashTable_t       *ip6_table;
    smFieldMap_t        *map;
    FILE                *out_file;
    char                *file_prefix;
    char                *last_file;
    md_dedup_tmpl_t     *tmpl;
    yfTime_t             last_rotate;
    uint64_t             count;
    yfTime_t             last_flush;
    /* 1 for SIP, 0 for DIP */
    int                  sip;
    /* TRUE if this is an ssl table */
    gboolean             ssl;
};


static fbInfoElementSpec_t  md_dedup_spec_add[] = {
    /* Millisecond first seen and last seen (epoch) (native time) */
    {C("monitoringIntervalStartMilliSeconds"), 8,  0},
    {C("monitoringIntervalEndMilliSeconds"),   8,  0},
    {C("flowStartMilliseconds"),               8,  0},
    {C("flowStartNanoseconds"),                8,  0},
    {C("observedDataTotalCount"),              8,  0},
    {C("sourceIPv6Address"),                   16, 0},
    {C("sourceIPv4Address"),                   4,  0},
    {C("yafFlowKeyHash"),                      4,  0},
    {C("observationDomainName"),               FB_IE_VARLEN, 0},
    FB_IESPEC_NULL
};


void
md_dedup_basic_list(
    fbBasicList_t  *bl,
    mdBuf_t        *buf,
    GString        *tstr,
    char            delim,
    gboolean        hex,
    gboolean        escape)
{
    uint16_t      k = 0;
    fbVarfield_t *var = NULL;
    fbVarfield_t *varlist[100];
    int           hits[100];
    int           varnum = 0;
    int           w = 0;
    gboolean      found;

    if (bl->numElements < 2) {
        /* not exciting - just add hit count and done */
        g_string_append_printf(tstr, "1%c", delim);
        mdPrintBasicList(buf, tstr, bl, delim, hex, escape);
        return;
    }

    memset(hits, 0, sizeof(hits));

    varlist[varnum] = (fbVarfield_t *)FBBLNEXT(bl, 0);
    hits[varnum] = 1;
    varnum++;

    for (k = 1; (var = (fbVarfield_t *)FBBLNEXT(bl, k)); k++) {
        found = FALSE;

        if (var->len == 0) {
            continue;
        }

        for (w = 0; w < varnum; w++) {
            if (var->len != varlist[w]->len) {
                continue;
            } else {
                if (memcmp(var->buf, varlist[w]->buf, var->len) == 0) {
                    hits[w]++;
                    found = TRUE;
                    break;
                }
            }
        }
        if (!found) {
            varlist[varnum] = var;
            hits[varnum] = 1;
            varnum++;
        }
    }

    for (k = 0; k < varnum; k++) {
        mdBufAppendGString(buf, tstr);
        mdBufAppendPrintf(buf, "%d%c", hits[k], delim);
        if (hex) {
            mdBufAppendHexdump(buf, varlist[k]->buf, varlist[k]->len, FALSE);
            /* no new-line here (2024.03.18: why?) */
        } else if (escape) {
            mdPrintEscapeChars(buf, varlist[k]->buf, varlist[k]->len, delim);
            mdBufAppendChar(buf, '\n');
        } else {
            mdBufAppendVarfield(buf, varlist[k]);
            mdBufAppendChar(buf, '\n');
        }
    }
}

void
md_dedup_basic_list_no_count(
    fbBasicList_t  *bl,
    mdBuf_t        *buf,
    char            delim,
    gboolean        quote,
    gboolean        hex,
    gboolean        escape)
{
    unsigned int  k;
    fbVarfield_t *var = NULL;
    fbVarfield_t *varlist[100];
    unsigned int  varnum = 0;
    unsigned int  w;
    gboolean      found;

    var = (fbVarfield_t *)FBBLNEXT(bl, 0);
    if (NULL == var) {
        return;
    }

    varlist[varnum] = var;
    varnum++;

    for (k = 1; (var = (fbVarfield_t *)FBBLNEXT(bl, k)); k++) {
        if (var->len == 0) {
            continue;
        }

        found = FALSE;
        for (w = 0; w < varnum; w++) {
            if ((var->len == varlist[w]->len) &&
                (memcmp(var->buf, varlist[w]->buf, var->len) == 0))
            {
                found = TRUE;
                break;
            }
        }
        if (!found) {
            varlist[varnum] = var;
            varnum++;
            if ((sizeof(varlist)/sizeof(varlist[0])) == varnum) {
                break;
            }
        }
    }

    if (quote) {
        mdBufAppendChar(buf, '\"');
    }

    /* there is always at least one item in the varlist */
    for (k = 0; k < varnum; k++) {
        if (hex) {
            mdBufAppendHexdump(buf, varlist[k]->buf, varlist[k]->len, FALSE);
        } else if (escape) {
            mdPrintEscapeChars(buf, varlist[k]->buf, varlist[k]->len,
                               (quote ? '"' : delim));
        } else {
            mdBufAppendVarfield(buf, varlist[k]);
        }
        if (quote) {
            /* close quote, delimit, open quote for next item */
            mdBufAppendPrintf(buf, "\"%c\"", delim);
        } else {
            mdBufAppendChar(buf, delim);
        }
    }

    /* remove final delimiter and unused opening quote */
    mdBufChop(buf, (quote ? 2 : 1));
}

void
md_dedup_print_stats(
    md_dedup_state_t  *state,
    char              *exp_name)
{
    if (state->stats.recvd == 0) {
        g_message("Exporter %s: %" PRIu64 " Records, %" PRIu64 " flushed",
                  exp_name, state->stats.recvd, state->stats.flushed);
        return;
    }

    g_message("Exporter %s: %" PRIu64 " Records, %" PRIu64 " flushed"
              "(%2.2f%% compression)",
              exp_name, state->stats.recvd, state->stats.flushed,
              (100.0 * (1.0 - (((double)state->stats.flushed) /
                               ((double)state->stats.recvd)))));
}

static void
md_dedup_ssl_decrement_cert(
    md_dedup_state_t         *state,
    md_dedup_ssl_str_node_t  *node)
{
    char            temp[4092];
    smVarHashKey_t  lookup;

    --(node->cert1->count);

    if (node->cert1->count == 0) {
        if ((node->cert1->issuer_len + node->cert1->serial_len) < 4092) {
            memcpy(temp, node->cert1->serial, node->cert1->serial_len);
            memcpy(temp + node->cert1->serial_len, node->cert1->issuer,
                   node->cert1->issuer_len);
            lookup.val = (uint8_t *)temp;
            lookup.len = node->cert1->serial_len + node->cert1->issuer_len;
            g_hash_table_remove(state->cert_table, &lookup);
            g_slice_free1(node->cert1->issuer_len, node->cert1->issuer);
            g_slice_free1(node->cert1->serial_len, node->cert1->serial);
            g_slice_free(md_dedup_ssl_node_t, node->cert1);
        }
    }

    if (node->cert2) {
        --(node->cert2->count);
        if (node->cert2->count == 0) {
            if ((node->cert2->issuer_len + node->cert2->serial_len) < 4092) {
                memcpy(temp, node->cert2->serial, node->cert2->serial_len);
                memcpy(temp + node->cert2->serial_len, node->cert2->issuer,
                       node->cert2->issuer_len);
                lookup.val = (uint8_t *)temp;
                lookup.len = node->cert2->serial_len + node->cert2->issuer_len;
                g_hash_table_remove(state->cert_table, &lookup);
                g_slice_free1(node->cert2->issuer_len, node->cert2->issuer);
                g_slice_free1(node->cert2->serial_len, node->cert2->serial);
                g_slice_free(md_dedup_ssl_node_t, node->cert2);
            }
        }
    }
}



gboolean
md_dedup_flush_queue(
    md_export_node_t  *exp,
    mdConfig_t        *cfg,
    GError           **err)
{
    md_dedup_node_t   *node = NULL;
    md_dedup_state_t  *state = exp->dedup;
    md_dedup_cqueue_t *cq = state->cq;

    if (cq == NULL) {
        return TRUE;
    }

    while ((node = detachFromEndOfDLL((mdDLL_t **)&(cq->head),
                                      (mdDLL_t **)&(cq->tail))))
    {
        if (!mdExporterDedupFileOpen(cfg, exp->exp, &(node->ietab->out_file),
                                     &(node->ietab->last_file),
                                     node->ietab->file_prefix,
                                     &node->ietab->last_rotate))
        {
            return FALSE;
        }
#if DEDUP_DEBUG
        g_debug("flushing queue node: %p, node->ietab %p, node->strnode %p",
                node, node->ietab, node->strnode);
        g_debug("file->prefix %s, intid %02x, extid %02x",
                node->ietab->file_prefix,
                node->ietab->tmpl->intid, node->ietab->tmpl->extid);
#endif /* if DEDUP_DEBUG */
        if (state->add_export) {
            if (node->exnode.mapname.len == 0) {
                node->exnode.mapname.buf =
                    (uint8_t *)mdExporterGetName(exp->exp);
                node->exnode.mapname.len = strlen(mdExporterGetName(exp->exp));
            }
        }

        if (!mdExporterWriteDedupRecord(
                cfg, exp, node->ietab->out_file, &(node->exnode),
                node->ietab->file_prefix, node->ietab->tmpl->intid,
                node->ietab->tmpl->extid, err))
        {
            return FALSE;
        }

        if (!node->ietab->ssl) {
            g_slice_free1(node->strnode->caplen, node->strnode->data);
            g_slice_free(md_dedup_str_node_t, node->strnode);
        } else {
            md_dedup_ssl_decrement_cert(
                state, (md_dedup_ssl_str_node_t *)node->strnode);
            g_slice_free(md_dedup_ssl_str_node_t,
                         (md_dedup_ssl_str_node_t *)node->strnode);
            /* check count on certs... free certs if necessary */
        }
        g_slice_free(md_dedup_node_t, node);

        state->stats.flushed++;
    }

    return TRUE;
}



md_dedup_state_t *
md_dedup_new_dedup_state(
    void)
{
    md_dedup_state_t *state = g_slice_new0(md_dedup_state_t);

    state->ie_table = g_hash_table_new((GHashFunc)g_direct_hash,
                                       (GEqualFunc)g_direct_equal);

    state->cq = g_slice_new0(md_dedup_cqueue_t);

    /* set defaults */
    state->max_hit_count = MD_DEDUP_DEFAULT_HIT;
    yfDiffTimeFromSeconds(&state->flush_timeout, MD_DEDUP_DEFAULT_FLUSH);

#if DEDUP_DEBUG
    fprintf(stderr, "created new dedup state %p\n", state->ie_table);
#endif

    return state;
}

gboolean
md_dedup_add_templates(
    md_dedup_state_t  *state,
    fBuf_t            *fbuf,
    GError           **err)
{
    md_dedup_tmpl_t       *tnode  = NULL;
    fbSession_t           *session = fBufGetSession(fbuf);
    md_dedup_ie_t         *tn = NULL, *cn = NULL;
    fbInfoModel_t         *sm_model = mdInfoModel();
    const fbInfoElement_t *ie_to_add  = NULL;
    fbInfoElementSpec_t    md_dedup_ie_spec = FB_IESPEC_NULL;
    GString               *template_name;

    for (tn = state->head; tn; tn = cn) {
        cn = tn->next;
        tnode = tn->tmpl;

        template_name = g_string_new("md_dedup");
        /* really should have configureable PEN */
        ie_to_add = fbInfoModelGetElementByID(sm_model, tnode->ie, CERT_PEN);

        if (tn->ssl == FALSE) {
            tnode->tmpl = fbTemplateAlloc(sm_model);

            if (!fbTemplateAppendSpecArray(tnode->tmpl, md_dedup_spec_add,
                                           0xffffffff, err))
            {
                g_string_free(template_name, TRUE);
                return FALSE;
            }

            md_dedup_ie_spec.name = (char *)ie_to_add->ref.name;
            md_dedup_ie_spec.len_override = 0;
            md_dedup_ie_spec.flags = 0;

            if (!fbTemplateAppendSpec(tnode->tmpl, &md_dedup_ie_spec, 0, err)) {
                g_string_free(template_name, TRUE);
                return FALSE;
            }

            tnode->intid = fbSessionAddTemplate(session, TRUE, tnode->intid,
                                                tnode->tmpl, err);
            if (tnode->intid == 0) {
                g_string_free(template_name, TRUE);
                return FALSE;
            }

#if SM_ENABLE_METADATA_EXPORT
            g_string_append_printf(template_name, "_%s", md_dedup_ie_spec.name);
            tnode->extid = fbSessionAddTemplateWithMetadata(
                session, FALSE, tnode->extid, tnode->tmpl,
                template_name->str, NULL, err);
#else  /* if SM_ENABLE_METADATA_EXPORT */
            tnode->extid = fbSessionAddTemplate(session, FALSE, tnode->extid,
                                                tnode->tmpl, err);
#endif /* if SM_ENABLE_METADATA_EXPORT */
            if (tnode->extid == 0) {
                g_string_free(template_name, TRUE);
                return FALSE;
            }
        } else {
            tnode->intid = MD_DEDUP_FULL;
            tnode->extid = MD_DEDUP_FULL;
        }
        g_string_free(template_name, TRUE);
    }

    /* if (!fbSessionExportTemplates(session, err)) { */
    /*     return FALSE; */
    /* } */
    return TRUE;
}

static void
md_dedup_reset(
    md_export_node_t  *exp,
    const yfTime_t    *cur_time)
{
    g_warning("Potentially out of memory for deduplication."
              " Resetting all tables.");
    md_dedup_flush_alltab(exp, cur_time, TRUE);
}

void
md_dedup_configure_state(
    md_dedup_state_t  *state,
    int                max_hit_count,
    int                flush_timeout,
    gboolean           merge_truncated,
    gboolean           add_export)
{
    if (max_hit_count > 0) {
        state->max_hit_count = max_hit_count;
    }

    if (flush_timeout > 0) {
        yfDiffTimeFromSeconds(&state->flush_timeout, flush_timeout);
    }

    if (merge_truncated) {
        state->merge = merge_truncated;
    }

    if (add_export) {
        state->add_export = add_export;
    }
}

static md_dedup_str_node_t *
md_dedup_new_str_node(
    uint8_t         *data,
    size_t           caplen,
    const yfTime_t  *cur_time,
    uint32_t         hash,
    const yfTime_t  *stime)
{
    md_dedup_str_node_t *stn;

    stn = g_slice_new0(md_dedup_str_node_t);

    stn->ftime = *cur_time;
    stn->ltime = *cur_time;
    stn->hitcount = 1;
    stn->hash = hash;
    stn->stime = *stime;

    stn->data = g_slice_alloc0(caplen);
    memcpy(stn->data, data, caplen);
    stn->caplen = caplen;

    return stn;
}

static md_dedup_ie_t *
md_dedup_ie_lookup(
    md_dedup_state_t  *state,
    uint16_t           ie)
{
    md_dedup_ie_t *ret = NULL;

    ret = g_hash_table_lookup(state->ie_table,
                              GUINT_TO_POINTER((unsigned int)ie));

    return ret;
}

void
md_dedup_add_ie(
    md_dedup_state_t  *state,
    md_dedup_ie_t     *ie_tab,
    uint16_t           ie)
{
#if DEDUP_DEBUG
    fprintf(stderr, "add ie %d to ietab %p\n", ie, ie_tab);
#endif
    g_hash_table_insert(state->ie_table,
                        GUINT_TO_POINTER((unsigned int)ie), ie_tab);
}


md_dedup_ie_t *
md_dedup_add_ie_table(
    md_dedup_state_t  *state,
    char              *prefix,
    smFieldMap_t      *map,
    uint16_t           ie,
    int                sip)
{
    md_dedup_ie_t   *ie_tab = NULL;
    md_dedup_tmpl_t *dedup_tmpl = NULL;

#if DEDUP_DEBUG
    fprintf(stderr, "state->ie_table is %p\n", state->ie_table);
#endif

    if ((ie_tab = g_hash_table_lookup(state->ie_table,
                                      GUINT_TO_POINTER((unsigned int)ie))))
    {
        /* already exists */
        return NULL;
    }

    ie_tab = g_slice_new0(md_dedup_ie_t);

    if (map) {
        ie_tab->ip_table = smCreateHashTable(
            sizeof(uint32_t) + sizeof(uint32_t), NULL, NULL);
        ie_tab->ip6_table = smCreateHashTable(
            sizeof(uint32_t) + sizeof(uint8_t[16]), NULL, NULL);
        /* ie_tab->ip_table = g_hash_table_new((GHashFunc)g_direct_hash, */
        /*                                     (GEqualFunc)g_direct_equal); */
        /* ie_tab->ip6_table = g_hash_table_new_full( */
        /*     (GHashFunc)sm_octet_array_hash, */
        /*     (GEqualFunc)sm_octet_array_equal, */
        /*     sm_octet_array_key_destroy, NULL); */
    } else {
        ie_tab->ip_table = smCreateHashTable(sizeof(uint32_t), NULL, NULL);
        ie_tab->ip6_table = smCreateHashTable(sizeof(uint8_t[16]), NULL, NULL);
    }

    ie_tab->file_prefix = g_strdup(prefix);

    ie_tab->sip = sip;

    ie_tab->map = map;

    g_hash_table_insert(state->ie_table, GUINT_TO_POINTER((unsigned int)ie),
                        ie_tab);

    attachHeadToDLL((mdDLL_t **)&(state->head),
                    (mdDLL_t **)&(state->tail),
                    (mdDLL_t *)ie_tab);

    /* if ie == serial #, then set up SSL state */
    if (ie == SSL_SERIAL_IE) {
        state->cert_table = g_hash_table_new_full(
            (GHashFunc)sm_octet_array_hash, (GEqualFunc)sm_octet_array_equal,
            sm_octet_array_key_destroy, NULL);
        ie_tab->ssl = TRUE;
    }

    dedup_tmpl = g_slice_new0(md_dedup_tmpl_t);

    dedup_tmpl->ie = ie;

    ie_tab->tmpl = dedup_tmpl;

    return ie_tab;
}

static void
md_dedup_ip_node_close(
    md_dedup_ie_t       *ietab,
    md_dedup_ip_node_t  *ipnode)
{
    if (ipnode->sip6_key) {
        /*g_hash_table_remove(ietab->ip6_table, ipnode->sip6_key);*/
        smHashTableRemove(ietab->ip6_table, (uint8_t *)(ipnode->sip6_key));
        g_slice_free(mdMapKey6_t, ipnode->sip6_key);
    } else {
        smHashTableRemove(ietab->ip_table, (uint8_t *)(ipnode->sip_key));
        g_slice_free(mdMapKey4_t, ipnode->sip_key);
        /* g_hash_table_remove(ietab->ip_table, */
        /*                     GUINT_TO_POINTER(ipnode->sip_key)); */
    }

    detachThisEntryOfDLL((mdDLL_t **)&(ietab->head),
                         (mdDLL_t **)&(ietab->tail),
                         (mdDLL_t *)ipnode);
#if DEDUP_DEBUG
    g_debug("REMOVE IPNODE %u", ipnode->sip_key);
#endif

    if (ietab->ssl) {
        g_slice_free(md_dedup_ssl_ip_node_t, (md_dedup_ssl_ip_node_t *)ipnode);
    } else {
        g_slice_free(md_dedup_ip_node_t, ipnode);
    }
    --(ietab->count);
}

static void
md_dedup_str_node_close(
    md_export_node_t     *exp,
    md_dedup_ie_t        *ietab,
    md_dedup_ip_node_t   *ipnode,
    md_dedup_str_node_t  *strnode)
{
    md_dedup_cqueue_t *cq = exp->dedup->cq;
    md_dedup_node_t   *cn = g_slice_new0(md_dedup_node_t);

    if (cn == NULL) {
        g_warning("Error allocating md_dedup_node.");
        return;
    }
#if DEDUP_DEBUG
    fprintf(stderr, "CLOSING STRNODE %p\n", strnode);
#endif
    cn->strnode = strnode;
    cn->ietab = ietab;
    cn->exnode.monitoringIntervalStartMilliSeconds
        = yfTimeToMilli(strnode->ftime);
    cn->exnode.monitoringIntervalEndMilliSeconds
        = yfTimeToMilli(strnode->ltime);
    cn->exnode.observedDataTotalCount = strnode->hitcount;
    cn->exnode.yafFlowKeyHash = strnode->hash;
    cn->exnode.flowStartMilliseconds = yfTimeToMilli(strnode->stime);
    yfTimeToNTP(&cn->exnode.flowStartNanoseconds, strnode->stime);
    if (ipnode->sip_key) {
        cn->exnode.sourceIPv4Address = ipnode->sip_key->ip;
    }
    if (ipnode->sip6_key) {
        /*         memcpy(cn->exnode.sip6, ipnode->sip6_key->val, 16); */
        memcpy(cn->exnode.sourceIPv6Address, ipnode->sip6_key, 16);
    }

    if (ietab->map) {
        int  mapindex = 0;
        if (ipnode->sip_key) {
            mapindex = ipnode->sip_key->map;
        } else {
            /*mapindex = ((mdMapKey6_t*)(ipnode->sip6_key->val))->map;*/
            mapindex = ((mdMapKey6_t *)(ipnode->sip6_key))->map;
        }
#if DEDUP_DEBUG
        g_debug("maps on %s", ietab->map->labels[mapindex]);
#endif
        cn->exnode.mapname.buf = (uint8_t *)(ietab->map->labels[mapindex]);
        cn->exnode.mapname.len = strlen(ietab->map->labels[mapindex]);
    }

    if (!ietab->ssl) {
        cn->exnode.observedData.buf = strnode->data;
        cn->exnode.observedData.len = strnode->caplen;
    } else {
        md_dedup_ssl_str_node_t *ssl = (md_dedup_ssl_str_node_t *)strnode;
        cn->exnode.sslCertSerialNumber1.buf = ssl->cert1->serial;
        cn->exnode.sslCertSerialNumber1.len = ssl->cert1->serial_len;
        cn->exnode.sslCertIssuerCommonName1.buf = ssl->cert1->issuer;
        cn->exnode.sslCertIssuerCommonName1.len = ssl->cert1->issuer_len;
        if (ssl->cert2) {
            cn->exnode.sslCertSerialNumber2.buf = ssl->cert2->serial;
            cn->exnode.sslCertSerialNumber2.len = ssl->cert2->serial_len;
            cn->exnode.sslCertIssuerCommonName2.buf = ssl->cert2->issuer;
            cn->exnode.sslCertIssuerCommonName2.len = ssl->cert2->issuer_len;
        }
    }

    attachHeadToDLL((mdDLL_t **)&(cq->head),
                    (mdDLL_t **)&(cq->tail),
                    (mdDLL_t *)cn);

    detachThisEntryOfDLL((mdDLL_t **)&(ipnode->head),
                         (mdDLL_t **)&(ipnode->tail),
                         (mdDLL_t *)strnode);

    if (!ipnode->head) {
        md_dedup_ip_node_close(ietab, ipnode);
    }
}


static void
md_dedup_str_node_tick(
    md_export_node_t     *exp,
    md_dedup_ie_t        *ietab,
    md_dedup_ip_node_t   *ipnode,
    md_dedup_str_node_t  *strnode)
{
    if (ipnode->head != strnode) {
        if (strnode->prev != NULL) {
            detachThisEntryOfDLL((mdDLL_t **)&(ipnode->head),
                                 (mdDLL_t **)&(ipnode->tail),
                                 (mdDLL_t *)strnode);
        }
        attachHeadToDLL((mdDLL_t **)&(ipnode->head),
                        (mdDLL_t **)&(ipnode->tail),
                        (mdDLL_t *)strnode);
    }

    while (ipnode->tail &&
           yfTimeCheckElapsed(strnode->ltime, ipnode->tail->ltime,
                              exp->dedup->flush_timeout))
    {
        md_dedup_str_node_close(exp, ietab, ipnode, ipnode->tail);
    }
}

static void
md_dedup_ip_node_tick(
    md_dedup_ie_t       *ietab,
    md_dedup_ip_node_t  *ipnode)
{
    if (ietab->head != ipnode) {
        if (ipnode->prev != NULL) {
            detachThisEntryOfDLL((mdDLL_t **)&(ietab->head),
                                 (mdDLL_t **)&(ietab->tail), (mdDLL_t *)ipnode);
        }

        attachHeadToDLL((mdDLL_t **)&(ietab->head),
                        (mdDLL_t **)&(ietab->tail),
                        (mdDLL_t *)ipnode);
    }
}

static void
md_dedup_add_node(
    mdContext_t       *ctx,
    md_export_node_t  *exp,
    md_dedup_ie_t     *ietab,
    uint8_t           *data,
    size_t             datalen,
    uint16_t           ie,
    mdFullFlow_t      *flow,
    gboolean           rev)
{
    md_dedup_state_t    *state = exp->dedup;
    md_dedup_ip_node_t  *ipnode = NULL;
    md_dedup_str_node_t *strnode = NULL,  *cn = NULL, *tn = NULL;
    size_t               cmpsize = datalen;
    int                  rc;
    uint32_t             sip = 0;
    uint32_t             hash = md_util_flow_key_hash(flow->rec);
    mdMapKey4_t          mapkey4;
    mdMapKey6_t          mapkey6;
    gboolean             v6 = FALSE;

    mapkey4.map = 0;
    mapkey6.map = 0;

    if (datalen == 0 || !data) {
        /* no data to add */
        return;
    }

    if (ietab->sip == 2) {
        mapkey4.ip = hash;
        sip = hash;
    } else {
        if (flow->rec->sourceIPv4Address || flow->rec->destinationIPv4Address) {
            if (rev || (!rev && !ietab->sip)) {
                mapkey4.ip = flow->rec->destinationIPv4Address;
                sip = flow->rec->destinationIPv4Address;
            } else {
                mapkey4.ip = flow->rec->sourceIPv4Address;
                sip = flow->rec->sourceIPv4Address;
            }
        } else {
            if (rev || (!rev && !ietab->sip)) {
                memcpy(mapkey6.ip, flow->rec->destinationIPv6Address, 16);
            } else {
                memcpy(mapkey6.ip, flow->rec->sourceIPv6Address, 16);
            }
            v6 = TRUE;
        }
    }

    if (ietab->map) {
        mapkey4.map = smFieldMapTranslate(ietab->map, flow);
        if (ietab->map->discard && (mapkey4.map == 0)) {
            return;
        }
        mapkey6.map = mapkey4.map;
    }

    if (v6) {
        ipnode = smHashLookup(ietab->ip6_table, (uint8_t *)&mapkey6);
    } else {
        ipnode = smHashLookup(ietab->ip_table, (uint8_t *)&mapkey4);
#if DEDUP_DEBUG
        g_debug("looking up sip %08x %p %04x - returned %p",
                mapkey4, ietab->ip_table, mapkey4.ip, ipnode);
#endif
    }

    if (ipnode) {
        for (tn = ipnode->head; tn; tn = cn) {
            cn = tn->next;
            if (ie != tn->ie) {
                continue;
            }
            if (!state->merge) {
                /* not merging truncated fields - so if lengths don't match,
                 * continue */
                if (datalen != tn->caplen) {
                    continue;
                }
            } else {
                cmpsize = datalen < tn->caplen ? datalen : tn->caplen;
            }
            rc = memcmp(tn->data, data, cmpsize);
            if (!rc) {
                state->stats.recvd++;
                ++(tn->hitcount);
                tn->hash = hash;
                tn->stime = flow->stime;
                tn->ltime = ctx->cfg->ctime;
                if (tn->hitcount == state->max_hit_count) {
                    md_dedup_str_node_close(exp, ietab, ipnode, tn);
                } else {
                    md_dedup_str_node_tick(exp, ietab, ipnode, tn);
                    md_dedup_ip_node_tick(ietab, ipnode);
                }
                return;
            }
        }
    } else {
        /* IP address not found for this IE */
        ipnode = g_slice_new0(md_dedup_ip_node_t);
        if (ipnode == NULL) {
            md_dedup_reset(exp, &ctx->cfg->ctime);
            return;
        }
        if (v6) {
            ipnode->sip6_key = g_slice_new0(mdMapKey6_t);
            memcpy(ipnode->sip6_key, &mapkey6, sizeof(mdMapKey6_t));
            smHashTableInsert(ietab->ip6_table, (uint8_t *)ipnode->sip6_key,
                              (uint8_t *)ipnode);
        } else {
            ipnode->sip_key = g_slice_new0(mdMapKey4_t);
            ipnode->sip_key->ip = sip;
            ipnode->sip_key->map = mapkey4.map;
            smHashTableInsert(ietab->ip_table, (uint8_t *)ipnode->sip_key,
                              (uint8_t *)ipnode);
            /* g_hash_table_insert(ietab->ip_table, */
            /*                     GUINT_TO_POINTER((unsigned int)sip), */
            /*                     ipnode); */
        }
        ++(ietab->count);
    }

    strnode = md_dedup_new_str_node(data, datalen, &ctx->cfg->ctime, hash,
                                    &flow->stime);

    if (strnode == NULL) {
        md_dedup_reset(exp, &ctx->cfg->ctime);
        return;
    }

    strnode->ie = ie;

    /* add to stats recvd count */
    state->stats.recvd++;

    md_dedup_str_node_tick(exp, ietab, ipnode, strnode);
    md_dedup_ip_node_tick(ietab, ipnode);
}

static void
md_dedup_free_ietab(
    md_export_node_t  *exp,
    md_dedup_ie_t     *ietab)
{
    /* g_hash_table_destroy(ietab->ip_table); */
    /* g_hash_table_destroy(ietab->ip6_table); */
    smHashTableFree(ietab->ip_table);
    smHashTableFree(ietab->ip6_table);

    if (ietab->out_file) {
        mdExporterDedupFileClose(exp->exp, ietab->out_file, ietab->last_file);
    }

    if (ietab->file_prefix) {
        g_free(ietab->file_prefix);
    }

    if (ietab->tmpl) {
        /* ietab->tmpl->tmpl should get freed when session is freed */
        g_slice_free(md_dedup_tmpl_t, ietab->tmpl);
    }

    g_slice_free(md_dedup_ie_t, ietab);
}

static void
md_dedup_flush_ietab(
    md_export_node_t  *exp,
    md_dedup_ie_t     *ietab,
    const yfTime_t    *cur_time,
    gboolean           flush_all)
{
    if (ietab == NULL) {
        return;
    }

    ietab->last_flush = *cur_time;

    while (flush_all && ietab->tail) {
        md_dedup_str_node_close(exp, ietab, ietab->tail, ietab->tail->tail);
    }

    while (ietab->tail &&
           yfTimeCheckElapsed(ietab->last_flush, ietab->tail->tail->ltime,
                              exp->dedup->flush_timeout))
    {
        md_dedup_str_node_close(exp, ietab, ietab->tail, ietab->tail->tail);
    }
}

void
md_dedup_flush_alltab(
    md_export_node_t  *exp,
    const yfTime_t    *cur_time,
    gboolean           flush_all)
{
    md_dedup_state_t *state = exp->dedup;
    md_dedup_ie_t    *tn = NULL, *cn = NULL;

    for (tn = state->head; tn; tn = cn) {
        cn = tn->next;
        md_dedup_flush_ietab(exp, tn, cur_time, flush_all);
    }
}

gboolean
md_dedup_free_state(
    mdConfig_t        *cfg,
    md_export_node_t  *exp,
    GError           **err)
{
    md_dedup_state_t *state = exp->dedup;
    md_dedup_ie_t    *tn = NULL, *cn = NULL;

    md_dedup_flush_alltab(exp, &cfg->ctime, TRUE);

    if (!md_dedup_flush_queue(exp, cfg, err)) {
        return FALSE;
    }

    for (tn = state->head; tn; tn = cn) {
        cn = tn->next;
        md_dedup_free_ietab(exp, tn);
    }

    g_hash_table_destroy(state->ie_table);

    if (state->cert_table) {
        g_hash_table_destroy(state->cert_table);
    }

    g_slice_free1(sizeof(md_dedup_cqueue_t), state->cq);

    return TRUE;
}


static void
md_dedup_add_bl(
    mdContext_t       *ctx,
    md_export_node_t  *exp,
    md_dedup_ie_t     *ietab,
    fbBasicList_t     *bl,
    mdFullFlow_t      *flow)
{
    uint16_t      w = 0;
    fbVarfield_t *var = NULL;

    for (w = 0;
         (var = (fbVarfield_t *)fbBasicListGetIndexedDataPtr(bl, w));
         w++)
    {
        if (var->len == 0) {
            continue;
        }

        md_dedup_add_node(ctx, exp, ietab, var->buf, var->len,
                          bl->infoElement->num, flow, FALSE);
    }
}

static md_dedup_ssl_node_t *
md_dedup_ssl_node(
    uint8_t  *serial,
    size_t    serial_len,
    uint8_t  *issuer,
    size_t    issuer_len)
{
    md_dedup_ssl_node_t *node = g_slice_new0(md_dedup_ssl_node_t);

    node->serial = g_slice_alloc0(serial_len);
    memcpy(node->serial, serial, serial_len);
    node->serial_len = serial_len;

    node->issuer = g_slice_alloc0(issuer_len);
    memcpy(node->issuer, issuer, issuer_len);
    node->issuer_len = issuer_len;

    return node;
}



static void
md_dedup_ssl_add_node(
    mdContext_t       *ctx,
    md_export_node_t  *exp,
    md_dedup_ie_t     *ietab,
    yaf_ssl_t         *ssl,
    mdFullFlow_t      *flow)
{
    yaf_ssl_cert_t          *cert = NULL;
    yaf_ssl_subcert_t       *obj = NULL;
    yaf_ssl_subcert_t       *ou = NULL;
    md_dedup_state_t        *state = exp->dedup;
    md_dedup_ssl_node_t     *cert1 = NULL, *cert2 = NULL;
    md_dedup_ssl_str_node_t *cn = NULL, *tn = NULL;
    md_dedup_ssl_str_node_t *strnode = NULL;
    md_dedup_ssl_ip_node_t  *ipnode = NULL;
    smVarHashKey_t           lookup;
    smVarHashKey_t          *newkey;
    uint32_t                 hash = md_util_flow_key_hash(flow->rec);
    uint32_t                 sip;
    mdMapKey4_t              mapkey4;
    mdMapKey6_t              mapkey6;
    gboolean                 found;
    gboolean                 v6 = FALSE;
    uint8_t                  temp[4092];
    int                      cert_no = 0;

    if (ietab->sip == 2) {
        mapkey4.ip = hash;
        sip = hash;
    } else {
        if (flow->rec->sourceIPv4Address || flow->rec->destinationIPv4Address) {
            if (ietab->sip == 0) {
                mapkey4.ip = flow->rec->destinationIPv4Address;
                sip = flow->rec->destinationIPv4Address;
            } else {
                mapkey4.ip = flow->rec->sourceIPv4Address;
                sip = flow->rec->sourceIPv4Address;
            }
        } else {
            if (ietab->sip == 0) {
                memcpy(mapkey6.ip, flow->rec->destinationIPv6Address, 16);
            } else {
                memcpy(mapkey6.ip, flow->rec->sourceIPv6Address, 16);
            }
            v6 = TRUE;
        }
    }

    if (ietab->map) {
        mapkey4.map = smFieldMapTranslate(ietab->map, flow);
        mapkey6.map = mapkey4.map;
    }

    while ((cert = (yaf_ssl_cert_t *)FBSTLNEXT(&(ssl->sslCertList), cert))) {
        obj = NULL;
        ou = NULL;
        if (cert->sslCertSerialNumber.len == 0) {
            /* no serial number */
            if (cert_no == 0) {
                return;
            } else {
                break;
            }
        }

        found = FALSE;
        while ((obj = (yaf_ssl_subcert_t *)FBSTLNEXT(&(cert->issuer), obj))) {
            if (obj->sslObjectType != SSL_COMMON_NAME) {
                if (obj->sslObjectType == SSL_ORG_UNIT) {
                    /* save just in case */
                    ou = obj;
                }
                continue;
            }

            if (obj->sslObjectValue.len == 0) {
                continue;
            }

            found = TRUE;
            break;
        }

        if (!found) {
            if (ou) {
                obj = ou;
            } else {
                return;
            }
        }

        if (cert->sslCertSerialNumber.len + obj->sslObjectValue.len < 4092) {
            memcpy(temp, cert->sslCertSerialNumber.buf,
                   cert->sslCertSerialNumber.len);
            memcpy(temp + cert->sslCertSerialNumber.len,
                   obj->sslObjectValue.buf,
                   obj->sslObjectValue.len);
        } else {
            /* cut this off somehow */
            g_debug("COMBO serial + issuer name over 4092");
            return;
        }

        lookup.val = temp;
        lookup.len = cert->sslCertSerialNumber.len + obj->sslObjectValue.len;

        if (!cert1) {
            cert1 = g_hash_table_lookup(state->cert_table, &lookup);

            if (!cert1) {
                /* add this cert */
                cert1 = md_dedup_ssl_node(cert->sslCertSerialNumber.buf,
                                          cert->sslCertSerialNumber.len,
                                          obj->sslObjectValue.buf,
                                          obj->sslObjectValue.len);
                newkey = sm_new_hash_key(lookup.val, lookup.len);
                g_hash_table_insert(state->cert_table, newkey, cert1);
            }
        } else if (!cert2) {
            cert2 = g_hash_table_lookup(state->cert_table, &lookup);

            if (!cert2) {
                /* add this cert */
                cert2 = md_dedup_ssl_node(cert->sslCertSerialNumber.buf,
                                          cert->sslCertSerialNumber.len,
                                          obj->sslObjectValue.buf,
                                          obj->sslObjectValue.len);
                newkey = sm_new_hash_key(lookup.val, lookup.len);
                g_hash_table_insert(state->cert_table, newkey, cert2);
            }
        }

        if (cert1 && cert2) {
            break;
        }
    }

    if (!cert1) {
        /* must have 1 valid cert! */
        return;
    }

    if (v6) {
        ipnode = smHashLookup(ietab->ip6_table, (uint8_t *)&mapkey6);
        /*ipnode = g_hash_table_lookup(ietab->ip6_table, &iplookup);*/
    } else {
        ipnode = smHashLookup(ietab->ip_table, (uint8_t *)&mapkey4);
#if DEDUP_DEBUG
        g_debug("looking up sip %u - returned %p", sip, ipnode);
#endif
        /* ipnode = g_hash_table_lookup( */
        /*     ietab->ip_table, GUINT_TO_POINTER((unsigned int)sip)); */
    }

    if (ipnode) {
        for (tn = ipnode->head; tn; tn = cn) {
            cn = tn->next;

            if (cert1 != tn->cert1) {
                continue;
            }
            if (cert2 != tn->cert2) {
                continue;
            }
            /* found a match */
            state->stats.recvd++;
            ++(tn->hitcount);
            tn->ltime = ctx->cfg->ctime;
            tn->hash = hash;
            tn->stime = flow->stime;
            if (tn->hitcount == state->max_hit_count) {
                md_dedup_str_node_close(exp, ietab,
                                        (md_dedup_ip_node_t *)ipnode,
                                        (md_dedup_str_node_t *)tn);
            } else {
                md_dedup_str_node_tick(exp, ietab, (md_dedup_ip_node_t *)ipnode,
                                       (md_dedup_str_node_t *)tn);
                md_dedup_ip_node_tick(ietab, (md_dedup_ip_node_t *)ipnode);
            }
            return;
        }
    } else {
        /* IP address not found in this table */
        ipnode = g_slice_new0(md_dedup_ssl_ip_node_t);
        if (ipnode == NULL) {
            md_dedup_reset(exp, &ctx->cfg->ctime);
            return;
        }

        if (v6) {
            ipnode->sip6_key = g_slice_new0(mdMapKey6_t);
            memcpy(ipnode->sip6_key, &mapkey6, sizeof(mdMapKey6_t));
            /* ipnode->sip6_key = sm_new_hash_key( */
            /*     iplookup.val, iplookup.len); */
            /* g_hash_table_insert( */
            /*     ietab->ip6_table, ipnode->sip6_key, ipnode); */
            smHashTableInsert(ietab->ip6_table, (uint8_t *)ipnode->sip6_key,
                              (uint8_t *)ipnode);
        } else {
            ipnode->sip_key = g_slice_new0(mdMapKey4_t);
            ipnode->sip_key->ip = sip;
            ipnode->sip_key->map = mapkey4.map;
            smHashTableInsert(ietab->ip_table, (uint8_t *)ipnode->sip_key,
                              (uint8_t *)ipnode);
            /* g_hash_table_insert(ietab->ip_table, */
            /*                     GUINT_TO_POINTER((unsigned int)sip), */
            /*                     ipnode); */
        }
        ++(ietab->count);
    }

    strnode = g_slice_new0(md_dedup_ssl_str_node_t);
    if (strnode == NULL) {
        md_dedup_reset(exp, &ctx->cfg->ctime);
        return;
    }

    strnode->ftime = ctx->cfg->ctime;
    strnode->ltime = ctx->cfg->ctime;
    strnode->hitcount = 1;
    strnode->hash = hash;
    strnode->stime = flow->stime;
    strnode->cert1 = cert1;
    strnode->cert2 = cert2;
    ++(cert1->count);
    if (cert2) {
        ++(cert2->count);
    }

    state->stats.recvd++;

    md_dedup_str_node_tick(exp, ietab, (md_dedup_ip_node_t *)ipnode,
                           (md_dedup_str_node_t *)strnode);
    md_dedup_ip_node_tick(ietab, (md_dedup_ip_node_t *)ipnode);
}

gboolean
md_dedup_write_dedup(
    mdContext_t       *ctx,
    md_export_node_t  *exp,
    md_dedup_t        *dedup,
    uint16_t           ie,
    GError           **err)
{
    md_dedup_state_t *state = exp->dedup;
    md_dedup_ie_t    *ietab = NULL;

    ietab = md_dedup_ie_lookup(state, ie);

    if (!ietab) {
        g_message("Ignoring incoming record: No IE dedup table for ie %d", ie);
        return TRUE;
    }

    if (!mdExporterDedupFileOpen(ctx->cfg, exp->exp, &(ietab->out_file),
                                 &(ietab->last_file),
                                 ietab->file_prefix,
                                 &ietab->last_rotate))
    {
        return FALSE;
    }

    if (!mdExporterWriteDedupRecord(ctx->cfg, exp, ietab->out_file,
                                    dedup, ietab->file_prefix,
                                    ietab->tmpl->intid, ietab->tmpl->extid,
                                    err))
    {
        return FALSE;
    }

    state->stats.flushed++;

    return TRUE;
}


void
md_dedup_lookup_node(
    mdContext_t       *ctx,
    md_export_node_t  *exp,
    mdFullFlow_t      *flow,
    GError           **err)
{
    md_dedup_state_t *state = exp->dedup;
    fbBasicList_t    *bl = NULL;
    md_dedup_ie_t    *ietab = NULL;
    gboolean          rev = FALSE;
    int               loop;

    MD_UNUSED_PARAM(err);

    if (flow->rec->reversePacketTotalCount) {
        rev = TRUE;
    }

    if (flow->p0f) {
        ietab = md_dedup_ie_lookup(state, 36);
        if (ietab) {
            md_dedup_add_node(ctx, exp, ietab, flow->p0f->osName.buf,
                              flow->p0f->osName.len, 36, flow, FALSE);
            if (rev) {
                md_dedup_add_node(ctx, exp, ietab, flow->p0f->reverseOsName.buf,
                                  flow->p0f->reverseOsName.len, 36, flow, TRUE);
            }
        }
        ietab = md_dedup_ie_lookup(state, 37);
        if (ietab) {
            md_dedup_add_node(ctx, exp, ietab, flow->p0f->osVersion.buf,
                              flow->p0f->osVersion.len, 37, flow, FALSE);
            if (rev) {
                md_dedup_add_node(
                    ctx, exp, ietab, flow->p0f->reverseOsVersion.buf,
                    flow->p0f->reverseOsVersion.len, 37, flow, TRUE);
            }
        }
        ietab = md_dedup_ie_lookup(state, 107);
        if (ietab) {
            md_dedup_add_node(ctx, exp, ietab, flow->p0f->osFingerPrint.buf,
                              flow->p0f->osFingerPrint.len, 107, flow, FALSE);
            if (rev) {
                md_dedup_add_node(
                    ctx, exp, ietab, flow->p0f->reverseOsFingerPrint.buf,
                    flow->p0f->reverseOsFingerPrint.len, 107, flow, TRUE);
            }
        }
    }

    if (flow->dhcpfpIndex) {
        fbSubTemplateMultiListEntry_t *entry;
        entry = fbSubTemplateMultiListGetIndexedEntry(
            &flow->rec->subTemplateMultiList, flow->dhcpfpIndex - 1);
        if ((entry->tmplID & YTF_BIF) == YAF_DHCP_FP_TID) {
            yaf_dhcp_fp_t *dhcp = NULL;
            ietab = md_dedup_ie_lookup(state, 242);
            if (ietab) {
                dhcp = (yaf_dhcp_fp_t *)FBSTMLNEXT(entry, dhcp);
                md_dedup_add_node(ctx, exp, ietab, dhcp->dhcpFingerPrint.buf,
                                  dhcp->dhcpFingerPrint.len, 242, flow, FALSE);
                if (entry->tmplID & YTF_REV) {
                    md_dedup_add_node(ctx, exp, ietab,
                                      dhcp->reverseDhcpFingerPrint.buf,
                                      dhcp->reverseDhcpFingerPrint.len, 242,
                                      flow, TRUE);
                }
            }
            ietab = md_dedup_ie_lookup(state, 243);
            if (ietab) {
                md_dedup_add_node(ctx, exp, ietab, dhcp->dhcpVendorCode.buf,
                                  dhcp->dhcpVendorCode.len, 243, flow, FALSE);
                if (entry->tmplID & YTF_REV) {
                    md_dedup_add_node(ctx, exp, ietab,
                                      dhcp->reverseDhcpVendorCode.buf,
                                      dhcp->reverseDhcpVendorCode.len, 243,
                                      flow, TRUE);
                }
            }
        } else if ((entry->tmplID & YTF_BIF) == YAF_DHCP_OPTIONS_TID) {
            yaf_dhcp_options_t *dhcp = NULL;
            ietab = md_dedup_ie_lookup(state, 243);
            if (ietab) {
                dhcp = (yaf_dhcp_options_t *)FBSTMLNEXT(entry, dhcp);
                md_dedup_add_node(ctx, exp, ietab, dhcp->dhcpVendorCode.buf,
                                  dhcp->dhcpVendorCode.len, 243, flow, FALSE);
                if (entry->tmplID & YTF_REV) {
                    md_dedup_add_node(ctx, exp, ietab,
                                      dhcp->reverseDhcpVendorCode.buf,
                                      dhcp->reverseDhcpVendorCode.len, 243,
                                      flow, TRUE);
                }
            }
        }
    }

    switch (flow->app_tid & YTF_BIF) {
      case YAF_HTTP_TID:
      case YAF_POP3_TID:
      case YAF_IRC_TID:
      case YAF_FTP_TID:
      case YAF_IMAP_TID:
      case YAF_SIP_TID:
      case YAF_RTSP_TID:
      case YAF_SSH_TID:
      case YAF_SMTP_211_TID:
      case YAF_NNTP_TID:
      case YAF_MODBUS_TID:
      case YAF_ENIP_TID:
        bl = (fbBasicList_t *)flow->app;
        for (loop = 0; loop < flow->app_elements; loop++) {
            if (bl->infoElement) {
                ietab = md_dedup_ie_lookup(state, bl->infoElement->num);
                if (ietab) {
                    md_dedup_add_bl(ctx, exp, ietab, bl, flow);
                }
            }
            bl++;
        }
        break;
      case YAF_TFTP_TID:
        {
            yaf_tftp_t *tftp = (yaf_tftp_t *)flow->app;
            ietab = md_dedup_ie_lookup(state, 127);
            if (ietab) {
                md_dedup_add_node(ctx, exp, ietab, tftp->tftpMode.buf,
                                  tftp->tftpMode.len, 127, flow, FALSE);
            }
            ietab = md_dedup_ie_lookup(state, 126);
            if (ietab) {
                md_dedup_add_node(ctx, exp, ietab, tftp->tftpFilename.buf,
                                  tftp->tftpFilename.len, 126, flow, FALSE);
            }
            break;
        }
      case YAF_SLP_TID:
        {
            yaf_slp_t *slp = (yaf_slp_t *)flow->app;
            char       slp_buffer[20];

            ietab = md_dedup_ie_lookup(state, 128);
            if (ietab) {
                snprintf(slp_buffer, sizeof(slp_buffer), "%d", slp->slpVersion);
                md_dedup_add_node(ctx, exp, ietab, (uint8_t *)slp_buffer,
                                  strlen(slp_buffer), 128, flow, FALSE);
            }
            ietab = md_dedup_ie_lookup(state, 129);
            if (ietab) {
                snprintf(slp_buffer, sizeof(slp_buffer), "%d",
                         slp->slpMessageType);
                md_dedup_add_node(ctx, exp, ietab, (uint8_t *)slp_buffer,
                                  strlen(slp_buffer), 129, flow, FALSE);
            }
            bl = (fbBasicList_t *)flow->app;
            if (bl->infoElement) {
                ietab = md_dedup_ie_lookup(state, bl->infoElement->num);
                if (ietab) {
                    md_dedup_add_bl(ctx, exp, ietab, bl, flow);
                }
            }
            break;
        }
      case YAF_SSL_TID:
        {
            yaf_ssl_t *sslflow = (yaf_ssl_t *)flow->app;

            ietab = md_dedup_ie_lookup(state, SSL_SERIAL_IE);
            if (!ietab) {
                break;
            }

            md_dedup_ssl_add_node(ctx, exp, ietab, sslflow, flow);
        }
        break;
      case YAF_MYSQL_TID:
        {
            yaf_mysql_t     *mflow = (yaf_mysql_t *)flow->app;
            yaf_mysql_txt_t *mtxt = NULL;
            ietab = md_dedup_ie_lookup(state, 223);
            if (ietab) {
                md_dedup_add_node(ctx, exp, ietab, mflow->mysqlUsername.buf,
                                  mflow->mysqlUsername.len, 223, flow, FALSE);
            }
            ietab = md_dedup_ie_lookup(state, 225);
            if (ietab) {
                while ((mtxt = ((yaf_mysql_txt_t *)
                                FBSTLNEXT(&(mflow->mysqlList), mtxt))))
                {
                    md_dedup_add_node(ctx, exp, ietab,
                                      mtxt->mysqlCommandText.buf,
                                      mtxt->mysqlCommandText.len, 225, flow,
                                      FALSE);
                }
            }
            break;
        }
      case YAF_DNS_TID:
        {
            yaf_dns_t   *dnsflow = (yaf_dns_t *)flow->app;
            yaf_dnsQR_t *dnsqrflow = NULL;

            /* dns query name */
            ietab = md_dedup_ie_lookup(state, 179);
            if (ietab) {
                while ((dnsqrflow = (yaf_dnsQR_t *)FBSTLNEXT(
                            &(dnsflow->dnsQRList), dnsqrflow)))
                {
                    /* just queries */
                    if (dnsqrflow->dnsQueryResponse == 0) {
                        md_dedup_add_node(ctx, exp, ietab,
                                          dnsqrflow->dnsQName.buf,
                                          dnsqrflow->dnsQName.len, 179, flow,
                                          FALSE);
                    }
                }
            }
        }
        break;
      case YAF_RTP_TID:
        {
            yaf_rtp_t *rtp = (yaf_rtp_t *)flow->app;
            char       rtp_buffer[20];

            ietab = md_dedup_ie_lookup(state, 287);
            if (ietab && rtp) {
                snprintf(rtp_buffer, sizeof(rtp_buffer), "%d",
                         rtp->rtpPayloadType);
                md_dedup_add_node(ctx, exp, ietab, (uint8_t *)rtp_buffer,
                                  strlen(rtp_buffer), 287, flow, FALSE);
                snprintf(rtp_buffer, sizeof(rtp_buffer), "%d",
                         rtp->reverseRtpPayloadType);
                /* for reverse fields, swap sip & dip */
                md_dedup_add_node(ctx, exp, ietab, (uint8_t *)rtp_buffer,
                                  strlen(rtp_buffer), 287, flow, TRUE);
            }
        }
        break;
      case YAF_DNP_TID:
        {
            yaf_dnp_t     *dnp = (yaf_dnp_t *)flow->app;
            yaf_dnp_rec_t *rec = NULL;
            char           dnp_buffer[65535];
            size_t         bufsz = sizeof(dnp_buffer);
            size_t         buflen;
            int            i;

            ietab = md_dedup_ie_lookup(state, 284);
            if (ietab && dnp) {
                while ((rec = (yaf_dnp_rec_t *)FBSTLNEXT(
                            &dnp->dnp_list, rec)))
                {
                    buflen = rec->dnp3ObjectData.len;
                    if (buflen > bufsz) {
                        buflen = bufsz;
                    }
                    i = md_util_hexdump_append(dnp_buffer, &bufsz,
                                               rec->dnp3ObjectData.buf, buflen);
                    md_dedup_add_node(ctx, exp, ietab, (uint8_t *)dnp_buffer,
                                      i, 284, flow, FALSE);
                }
            }
        }
        break;
      default:
        break;
    }

    /* attempt to flush all tables */
    md_dedup_flush_alltab(exp, &ctx->cfg->ctime, FALSE);
}
