/*
 *  Copyright 2012-2025 Carnegie Mellon University
 *  See license information in LICENSE.txt.
 */
/*
 *  mediator_filter.c
 *
 *  Functions to limit which IPFIX records are processed based on values
 *  within the record.
 *
 *  ------------------------------------------------------------------------
 *  Authors: Emily Sarneso
 *  ------------------------------------------------------------------------
 *  @DISTRIBUTION_STATEMENT_BEGIN@
 *  super_mediator-1.11
 *
 *  Copyright 2024 Carnegie Mellon University.
 *
 *  NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING
 *  INSTITUTE MATERIAL IS FURNISHED ON AN "AS-IS" BASIS. CARNEGIE MELLON
 *  UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED,
 *  AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR
 *  PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS OBTAINED FROM USE OF
 *  THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF
 *  ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT
 *  INFRINGEMENT.
 *
 *  Licensed under a GNU GPL 2.0-style license, please see LICENSE.txt or
 *  contact permission@sei.cmu.edu for full terms.
 *
 *  [DISTRIBUTION STATEMENT A] This material has been approved for public
 *  release and unlimited distribution.  Please see Copyright notice for
 *  non-US Government use and distribution.
 *
 *  This Software includes and/or makes use of Third-Party Software each
 *  subject to its own license.
 *
 *  DM24-1038
 *  @DISTRIBUTION_STATEMENT_END@
 *  ------------------------------------------------------------------------
 */

#include "mediator_filter.h"
#include "mediator_util.h"

#if ENABLE_SKIPSET
#include SKIPSET_HEADER_NAME
#if HAVE_SILK_SKIPADDR_H
#include <silk/skipaddr.h>
#endif
#endif /* if ENABLE_SKIPSET */



/**
 * mdComparison
 *
 * compare val_one to val_two with the given oper
 *
 * @param val_one - a value to be compared
 * @param val_two - the other value to be compared
 * @param oper - an operation such as '=', or '<'
 * @return TRUE - if the comparison is TRUE.
 *
 */
static gboolean
mdComparison(
    uint64_t        val_one,
    uint64_t        val_two,
    fieldOperator   oper)
{
    switch (oper) {
      case EQUAL:
        if (val_one == val_two) {
            return TRUE;
        }
        return FALSE;
      case NOT_EQUAL:
        if (val_one != val_two) {
            return TRUE;
        }
        return FALSE;
      case LESS_THAN:
        if (val_one < val_two) {
            return TRUE;
        }
        return FALSE;
      case LESS_THAN_OR_EQUAL:
        if (val_one <= val_two) {
            return TRUE;
        }
        return FALSE;
      case GREATER_THAN:
        if (val_one > val_two) {
            return TRUE;
        }
        return FALSE;
      case GREATER_THAN_OR_EQUAL:
        if (val_one >= val_two) {
            return TRUE;
        }
        return FALSE;
      default:
        return FALSE;
    }

    return FALSE;
}

#if ENABLE_SKIPSET
/**
 * mdCompareIPSet
 *
 * @param flow - the basic flow record
 * @param filter - the filter node
 *
 * @return TRUE if the filter passed
 *
 */
static gboolean
mdCompareIPSet(
    const md_main_template_t  *flow,
    const md_filter_t         *filter)
{
    skipaddr_t  addr;

    if (flow->sourceIPv4Address || flow->destinationIPv4Address) {
        switch (filter->field) {
          case SIP_ANY:
          case SIP_V4:
            skipaddrSetV4(&addr, &(flow->sourceIPv4Address));
            if (skIPSetCheckAddress(filter->ipset->ipset, &addr)) {
                return TRUE;
            }
            return FALSE;
          case DIP_ANY:
          case DIP_V4:
            skipaddrSetV4(&addr, &(flow->destinationIPv4Address));
            if (skIPSetCheckAddress(filter->ipset->ipset, &addr)) {
                return TRUE;
            }
            return FALSE;
          case ANY_IP:
            skipaddrSetV4(&addr, &(flow->sourceIPv4Address));
            if (skIPSetCheckAddress(filter->ipset->ipset, &addr)) {
                return TRUE;
            }
            skipaddrSetV4(&addr, &(flow->destinationIPv4Address));
            if (skIPSetCheckAddress(filter->ipset->ipset, &addr)) {
                return TRUE;
            }
            return FALSE;
          default:
            return FALSE;
        }
    }
#if HAVE_SKIPADDRSETV6
    else {
        switch (filter->field) {
          case SIP_ANY:
          case SIP_V6:
            skipaddrSetV6(&addr, flow->sourceIPv6Address);
            if (skIPSetCheckAddress(filter->ipset->ipset, &addr)) {
                return TRUE;
            }
            return FALSE;
          case DIP_ANY:
          case DIP_V6:
            skipaddrSetV6(&addr, flow->destinationIPv6Address);
            if (skIPSetCheckAddress(filter->ipset->ipset, &addr)) {
                return TRUE;
            }
            return FALSE;
          case ANY_IP:
          case ANY_IP6:
            skipaddrSetV6(&addr, flow->sourceIPv6Address);
            if (skIPSetCheckAddress(filter->ipset->ipset, &addr)) {
                return TRUE;
            }
            skipaddrSetV6(&addr, flow->destinationIPv6Address);
            if (skIPSetCheckAddress(filter->ipset->ipset, &addr)) {
                return TRUE;
            }
            return FALSE;
          default:
            return FALSE;
        }
    }
#endif  /* HAVE_SKIPADDRSETV6 */

    return FALSE;
}
#endif  /* ENABLE_SKIPSET */


/**
 * mdCollectionFilter
 *
 * loop through the filters and compare the filters expressions
 * to the given flow record.
 * only one needs to result to TRUE to pass and break out of loop
 *
 * @param filter - a list of filters
 * @param flow - the basic flow record
 * @param obdomain - the observation domain
 * @return TRUE if one of the filters passed
 */
static gboolean
mdCollectionFilter(
    const md_filter_t         *filter,
    const md_main_template_t  *flow,
    uint32_t                   obdomain,
    uint8_t                    collector_id)
{
    const md_filter_t *cfil;

    if (filter == NULL) {
        return TRUE;
    }

    for (cfil = filter; cfil != NULL; cfil = cfil->next) {
        /* stats record - only 2 fields apply */
        if (!flow) {
            if (cfil->field == OBDOMAIN) {
                if (mdComparison(obdomain, cfil->val[0], cfil->oper)) {
                    return TRUE;
                }
            } else if (cfil->field == COLLECTOR) {
                if (mdComparison(collector_id, cfil->val[0], cfil->oper)) {
                    return TRUE;
                }
            }
            continue;
        }

#if ENABLE_SKIPSET
        if (cfil->ipset) {
            if (cfil->oper == IN_LIST) {
                if (mdCompareIPSet(flow, cfil)) {
                    return TRUE;
                }
            } else if (cfil->oper == NOT_IN_LIST) {
                if (!mdCompareIPSet(flow, cfil)) {
                    return TRUE;
                }
            }
            continue;
        }
#endif  /* ENABLE_SKIPSET */

        switch (cfil->field) {
          case SIP_V4:
            if (mdComparison(flow->sourceIPv4Address, cfil->val[0],
                             cfil->oper))
            {
                return TRUE;
            }
            continue;
          case DIP_V4:
            if (mdComparison(flow->destinationIPv4Address, cfil->val[0],
                             cfil->oper))
            {
                return TRUE;
            }
            continue;
          case SPORT:
            if (mdComparison(flow->sourceTransportPort, cfil->val[0],
                             cfil->oper))
            {
                return TRUE;
            }
            continue;
          case DPORT:
            if (mdComparison(flow->destinationTransportPort, cfil->val[0],
                             cfil->oper))
            {
                return TRUE;
            }
            continue;
          case PROTOCOL:
            if (mdComparison(flow->protocolIdentifier, cfil->val[0],
                             cfil->oper))
            {
                return TRUE;
            }
            continue;
          case APPLICATION:
            if (mdComparison(flow->silkAppLabel, cfil->val[0], cfil->oper)) {
                return TRUE;
            }
            continue;
          case SIP_V6:
            if (memcmp(flow->sourceIPv6Address, cfil->val, 16) == 0) {
                return TRUE;
            }
            continue;
          case DIP_V6:
            if (memcmp(flow->destinationIPv6Address, cfil->val, 16) == 0) {
                return TRUE;
            }
            continue;
          case ANY_IP6:
            if (memcmp(flow->sourceIPv6Address, cfil->val, 16) != 0) {
                if (memcmp(flow->destinationIPv6Address, cfil->val, 16) != 0) {
                    continue;
                }
            }
            return TRUE;
          case ANY_IP:
            if (!mdComparison(flow->sourceIPv4Address, cfil->val[0],
                              cfil->oper))
            {
                if (!mdComparison(flow->destinationIPv4Address,
                                  cfil->val[0], cfil->oper))
                {
                    continue;
                }
            }
            return TRUE;
          case ANY_PORT:
            if (!mdComparison(flow->sourceTransportPort, cfil->val[0],
                              cfil->oper))
            {
                if (!mdComparison(flow->destinationTransportPort,
                                  cfil->val[0], cfil->oper))
                {
                    continue;
                }
            }
            return TRUE;
          case OBDOMAIN:
            if (mdComparison(obdomain, cfil->val[0], cfil->oper)) {
                return TRUE;
            }
            continue;
          case VLAN:
            if (mdComparison(flow->vlanId, cfil->val[0], cfil->oper)) {
                return TRUE;
            }
            continue;
          case IPVERSION:
            if (flow->sourceIPv4Address || flow->destinationIPv4Address) {
                if (cfil->val[0] == 4) {
                    return TRUE;
                } else if (cfil->val[0] == 6) {
                    continue;
                }
            } else {
                if (cfil->val[0] == 4) {
                    continue;
                } else if (cfil->val[0] == 6) {
                    return TRUE;
                }
            }
            continue;
          case INGRESS:
            if (mdComparison(flow->ingressInterface, cfil->val[0],
                             cfil->oper))
            {
                return TRUE;
            }
            continue;
          case EGRESS:
            if (mdComparison(flow->egressInterface, cfil->val[0], cfil->oper)) {
                return TRUE;
            }
            continue;
          case COLLECTOR:
            if (mdComparison(collector_id, cfil->val[0], cfil->oper)) {
                return TRUE;
            }
            continue;
          default:
            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING,
                  "BAD KEYWORD in Filter .. Ignoring\n");
            continue;
        }
    }

    return FALSE;
}

/**
 * mdCollectionANDFilter
 *
 * loop through the filters and compare the filters expressions
 * to the given flow record.
 * Only one filter needs to FAIL to break out of the loop.
 *
 * @param filter - a list of filters
 * @param flow - the basic flow record
 * @param obdomain - the observation domain
 * @return TRUE if one of the filters passed
 */
static gboolean
mdCollectionANDFilter(
    const md_filter_t         *filter,
    const md_main_template_t  *flow,
    uint32_t                   obdomain,
    uint8_t                    collector_id)
{
    const md_filter_t *cfil;

    if (filter == NULL) {
        return TRUE;
    }

    for (cfil = filter; cfil != NULL; cfil = cfil->next) {
        /* stats/dedup/dnsrr record - only 2 fields apply */
        if (!flow) {
            if (cfil->field == OBDOMAIN) {
                if (mdComparison(obdomain, cfil->val[0], cfil->oper)) {
                    continue;
                } else {
                    return FALSE;
                }
            } else if (cfil->field == COLLECTOR) {
                if (mdComparison(collector_id, cfil->val[0], cfil->oper)) {
                    continue;
                } else {
                    return FALSE;
                }
            }
            return FALSE;
        }

#if ENABLE_SKIPSET
        if (cfil->ipset) {
            if (cfil->oper == IN_LIST) {
                if (mdCompareIPSet(flow, cfil)) {
                    continue;
                }
            } else if (cfil->oper == NOT_IN_LIST) {
                if (!mdCompareIPSet(flow, cfil)) {
                    continue;
                }
            }
            return FALSE;
        }
#endif  /* ENABLE_SKIPSET */

        switch (cfil->field) {
          case SIP_V4:
            if (mdComparison(flow->sourceIPv4Address, cfil->val[0],
                             cfil->oper))
            {
                continue;
            }
            return FALSE;
          case DIP_V4:
            if (mdComparison(flow->destinationIPv4Address, cfil->val[0],
                             cfil->oper))
            {
                continue;
            }
            return FALSE;
          case SPORT:
            if (mdComparison(flow->sourceTransportPort, cfil->val[0],
                             cfil->oper))
            {
                continue;
            }
            return FALSE;
          case DPORT:
            if (mdComparison(flow->destinationTransportPort, cfil->val[0],
                             cfil->oper))
            {
                continue;
            }
            return FALSE;
          case PROTOCOL:
            if (mdComparison(flow->protocolIdentifier, cfil->val[0],
                             cfil->oper))
            {
                continue;
            }
            return FALSE;
          case APPLICATION:
            if (mdComparison(flow->silkAppLabel, cfil->val[0], cfil->oper)) {
                continue;
            }
            return FALSE;
          case SIP_V6:
            if (memcmp(flow->sourceIPv6Address, cfil->val, 16) == 0) {
                continue;
            }
            return FALSE;
          case DIP_V6:
            if (memcmp(flow->destinationIPv6Address, cfil->val, 16) == 0) {
                continue;
            }
            return FALSE;
          case ANY_IP6:
            if (memcmp(flow->sourceIPv6Address, cfil->val, 16) != 0) {
                if (memcmp(flow->destinationIPv6Address, cfil->val, 16) != 0) {
                    return FALSE;
                }
            }
            continue;
          case ANY_IP:
            if (!mdComparison(flow->sourceIPv4Address, cfil->val[0],
                              cfil->oper))
            {
                if (!mdComparison(flow->destinationIPv4Address,
                                  cfil->val[0], cfil->oper))
                {
                    return FALSE;
                }
            }
            continue;
          case ANY_PORT:
            if (!mdComparison(flow->sourceTransportPort, cfil->val[0],
                              cfil->oper))
            {
                if (!mdComparison(flow->destinationTransportPort,
                                  cfil->val[0], cfil->oper))
                {
                    return FALSE;
                }
            }
            continue;
          case OBDOMAIN:
            if (mdComparison(obdomain, cfil->val[0], cfil->oper)) {
                continue;
            }
            return FALSE;
          case VLAN:
            if (mdComparison(flow->vlanId, cfil->val[0], cfil->oper)) {
                continue;
            }
            return FALSE;
          case IPVERSION:
            if (flow->sourceIPv4Address || flow->destinationIPv4Address) {
                if (cfil->val[0] == 4) {
                    continue;
                } else if (cfil->val[0] == 6) {
                    return FALSE;
                }
            } else {
                if (cfil->val[0] == 4) {
                    return FALSE;
                } else if (cfil->val[0] == 6) {
                    continue;
                }
            }
            continue;
          case INGRESS:
            if (mdComparison(flow->ingressInterface, cfil->val[0],
                             cfil->oper))
            {
                continue;
            }
            return FALSE;
          case EGRESS:
            if (mdComparison(flow->egressInterface, cfil->val[0], cfil->oper)) {
                continue;
            }
            return FALSE;
          case COLLECTOR:
            if (mdComparison(collector_id, cfil->val[0], cfil->oper)) {
                continue;
            }
            return FALSE;
          default:
            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO,
                  "BAD KEYWORD in Filter .. Ignoring\n");
            continue;
        }
    }

    return TRUE;
}

/**
 * mdFilter
 *
 * @param filter - a list of filters
 * @param flow - the basic flow record
 * @param obdomain - the observation domain
 * @param and_filter - true if filters should be ANDed
 * @return TRUE if one of the filters passed
 *
 */
gboolean
mdFilter(
    const md_filter_t   *filter,
    const mdFullFlow_t  *md_flow,
    uint32_t             obdomain,
    gboolean             and_filter,
    uint8_t              collector_id)
{
    const md_main_template_t  *flow;
    gboolean  rc;

    flow = ((NULL == md_flow) ? NULL : md_flow->rec);

    if (and_filter) {
        rc = mdCollectionANDFilter(filter, flow, obdomain, collector_id);
    } else {
        rc = mdCollectionFilter(filter, flow, obdomain, collector_id);
    }

    return rc;
}

/**
 * md_new_filter_node
 *
 */
md_filter_t *
md_new_filter_node(
    void)
{
    md_filter_t *mf = g_slice_new0(md_filter_t);

    return mf;
}

void
md_free_filter_node(
    md_filter_t  *mf)
{
    if (mf) {
#if ENABLE_SKIPSET
        if (mf->ipset) {
            mdUtilIPSetClose(mf->ipset);
        }
#endif  /* ENABLE_SKIPSET */
        g_slice_free(md_filter_t, mf);
    }
}

#ifdef HAVE_SPREAD
/**
 * mdSpreadExporterFilter
 *
 * loop through spread filters and add the groups
 * that should receive this flow.  If a group does not
 * have a filter, it automatically gets all flows.
 *
 */
int
mdSpreadExporterFilter(
    const md_spread_filter_t  *sf,
    const mdFullFlow_t        *md_flow,
    char                     **groups)
{
    const md_spread_filter_t *cfil;
    int                       num_groups = 0;
    const md_main_template_t *flow = NULL;

    if (md_flow) {
        flow = md_flow->rec;
    }

    for (cfil = sf; cfil != NULL; cfil = cfil->next) {
        if (cfil->filterList == NULL) {
            /* there is no filter for this group so it gets everything */
            groups[num_groups] = cfil->group;
            num_groups++;
        } else {
            if (mdCollectionFilter(cfil->filterList, flow, 0, 0)) {
                if (num_groups < 10) {
                    groups[num_groups] = cfil->group;
                    num_groups++;
                }
            }
        }
    }

    return num_groups;
}

/**
 * md_new_spread_node
 *
 *
 */
md_spread_filter_t *
md_new_spread_node(
    void)
{
    md_spread_filter_t *ms = g_slice_new0(md_spread_filter_t);
    return ms;
}

void
md_free_spread_node(
    md_spread_filter_t  *ms)
{
    g_slice_free(md_spread_filter_t, ms);
}

#endif  /* HAVE_SPREAD */
